@@ -3,16 +3,23 @@
 """
 Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
-SPDX-License-Identifier: GPL-2.0
+SPDX-License-Identifier: GPL-2.0-only
 
 """
 
+from typing import TYPE_CHECKING, Sequence, Tuple, Union, Optional, Any, Dict, List, cast
+
 import weakref
+from itertools import chain
 
 from message import Message
-from common import IntDict, Debug, StrMixin
+from common import IntDict, Debug, ReprMixin
 from tile import Tile
 
+if TYPE_CHECKING:
+    from player import PlayingPlayer
+    from tile import Tiles, TileTuple, MeldList, Meld
+    from hand import Hand
 
 class AIDefaultAI:
 
@@ -20,24 +27,25 @@ class AIDefaultAI:
 
     groupPrefs = dict(zip(Tile.colors + Tile.honors, (0, 0, 0, 4, 7)))
 
-    # pylint: disable=no-self-use
     # we could solve this by moving those filters into DiscardCandidates
     # but that would make it more complicated to define alternative AIs
 
-    def __init__(self, player=None):
-        self._player = weakref.ref(player) if player else None
+    def __init__(self, player:'PlayingPlayer') ->None:
+        self._player = weakref.ref(player)
 
     @property
-    def player(self):
+    def player(self) ->'PlayingPlayer':
         """hide weakref"""
-        return self._player() if self._player else None  # pylint: disable=not-callable
+        result = self._player()
+        assert result
+        return result
 
-    def name(self):
+    def name(self) ->str:
         """return our name"""
         return self.__class__.__name__[2:]
 
     @staticmethod
-    def weighSameColors(unusedAiInstance, candidates):
+    def weighSameColors(unusedAiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
         """weigh tiles of same group against each other"""
         for candidate in candidates:
             if candidate.group in Tile.colors:
@@ -55,8 +63,7 @@ class AIDefaultAI:
                     candidate.next2.keep += 0.503
         return candidates
 
-    def selectDiscard(self, hand):
-        # pylint: disable=too-many-branches, R0915
+    def selectDiscard(self, hand:'Hand') ->Tile:
         # disable warning about too many branches
         """return exactly one tile for discard.
         Much of this is just trial and success - trying to get as much AI
@@ -67,40 +74,41 @@ class AIDefaultAI:
         candidates.unlink()
         return result
 
-    def weighDiscardCandidates(self, candidates):
+    def weighDiscardCandidates(self, candidates:'DiscardCandidates') ->'DiscardCandidates':
         """the standard"""
         game = self.player.game
+        assert game
         weighRules = game.ruleset.filterRules('weigh')
-        for aiFilter in [self.weighBasics, self.weighSameColors,
+        for _ in [self.weighBasics, self.weighSameColors,
                          self.weighSpecialGames, self.weighCallingHand,
                          self.weighOriginalCall,
                          self.alternativeFilter] + weighRules:
-            if aiFilter in weighRules:
-                filterName = aiFilter.__class__.__name__
-                aiFilter = aiFilter.weigh
+            if _ in weighRules:
+                filterName = _.__class__.__name__
+                aiFilter = _.weigh
             else:
-                filterName = aiFilter.__name__
+                filterName = _.__name__
+                aiFilter = _
             if Debug.robotAI:
                 prevWeights = ((x.tile, x.keep) for x in candidates)
                 candidates = aiFilter(self, candidates)
                 newWeights = ((x.tile, x.keep) for x in candidates)
                 for oldW, newW in zip(prevWeights, newWeights):
                     if oldW != newW:
-                        game.debug('%s: %s: %.3f->%.3f' % (
-                            filterName, oldW[0], oldW[1], newW[1]))
+                        game.debug(f'{filterName}: {oldW[0]}: {oldW[1]:.3f}->{newW[1]:.3f}')
             else:
                 candidates = aiFilter(self, candidates)
         return candidates
 
     @staticmethod
-    def alternativeFilter(unusedAiInstance, candidates):
+    def alternativeFilter(unusedAiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
         """if the alternative AI only adds tests without changing
         default filters, you can override this one to minimize
         the source size of the alternative AI"""
         return candidates
 
     @staticmethod
-    def weighBasics(aiInstance, candidates):
+    def weighBasics(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
         """basic things"""
         # pylint: disable=too-many-branches
         # too many branches
@@ -147,7 +155,7 @@ class AIDefaultAI:
         return candidates
 
     @staticmethod
-    def weighSpecialGames(unusedAiInstance, candidates):
+    def weighSpecialGames(unusedAiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
         """like color game, many dragons, many winds"""
         # pylint: disable=too-many-nested-blocks
         for candidate in candidates:
@@ -175,28 +183,31 @@ class AIDefaultAI:
         return candidates
 
     @staticmethod
-    def weighOriginalCall(aiInstance, candidates):
+    def weighOriginalCall(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
         """if we declared Original Call, respect it"""
         myself = aiInstance.player
         game = myself.game
+        assert game
         if myself.originalCall and myself.mayWin:
+            last = myself.lastTile
             if Debug.originalCall:
-                game.debug('weighOriginalCall: lastTile=%s, candidates=%s' %
-                           (myself.lastTile, [str(x) for x in candidates]))
+                game.debug(f'weighOriginalCall: lastTile={last}, candidates={[str(x) for x in candidates]}')
             for candidate in candidates:
-                if candidate.tile is myself.lastTile.exposed:
+                if last and last.exposed and candidate.tile:
+                    assert myself.originalCallingHand
                     winningTiles = myself.originalCallingHand.chancesToWin()
                     if Debug.originalCall:
-                        game.debug('weighOriginalCall: winningTiles=%s for %s' %
-                                   (winningTiles, str(myself.originalCallingHand)))
-                        game.debug('weighOriginalCall respects originalCall: %s with %d' %
-                                   (candidate.tile, -99 * len(winningTiles)))
+                        game.debug(f'weighOriginalCall: winningTiles={winningTiles} '
+                                   f'for {str(myself.originalCallingHand)}')
+                        game.debug(f'weighOriginalCall respects originalCall: '
+                                   f'{candidate.tile} with {int(-99 * len(winningTiles))}')
                     candidate.keep -= 99 * len(winningTiles)
         return candidates
 
     @staticmethod
-    def weighCallingHand(aiInstance, candidates):
+    def weighCallingHand(aiInstance:'AIDefaultAI', candidates:'DiscardCandidates') ->'DiscardCandidates':
         """if we can get a calling hand, prefer that"""
+        assert aiInstance.player.game
         for candidate in candidates:
             newHand = candidates.hand - candidate.tile.concealed
             winningTiles = newHand.chancesToWin()
@@ -204,28 +215,32 @@ class AIDefaultAI:
                 for winnerTile in sorted(set(winningTiles)):
                     winnerHand = newHand + winnerTile.concealed
                     if Debug.robotAI:
-                        aiInstance.player.game.debug('weighCallingHand %s cand %s winnerTile %s winnerHand %s: %s' % (
-                            newHand, candidate, winnerTile, winnerHand, '     '.join(winnerHand.explain())))
+                        aiInstance.player.game.debug(f"weighCallingHand {newHand} cand {candidate} "
+                                                     f"winnerTile {winnerTile} "
+                                                     f"winnerHand {winnerHand}: "
+                                                     f"{'     '.join(winnerHand.explain())}")
                     keep = winnerHand.total() / 10.017
                     candidate.keep -= keep
                     if Debug.robotAI:
                         aiInstance.player.game.debug(
-                            'weighCallingHand %s winnerTile %s: discardCandidate %s keep -= %.4f' % (
-                                newHand, winnerTile, candidate, keep))
+                            f'weighCallingHand {newHand} winnerTile {winnerTile}: '
+                            f'discardCandidate {candidate} keep -= {keep:.4f}')
                 # more weight if we have several chances to win
                 candidate.keep -= float(len(winningTiles)) / len(
                     set(winningTiles)) * 5.031
                 if Debug.robotAI:
-                    aiInstance.player.game.debug('weighCallingHand %s for %s winningTiles:%s' % (
-                        newHand, candidates.hand, winningTiles))
+                    assert aiInstance.player.game
+                    aiInstance.player.game.debug(f'weighCallingHand {newHand} for '
+                                                 f'{candidates.hand} winningTiles:{winningTiles}')
         return candidates
 
-    def selectAnswer(self, answers):
+    def selectAnswer(self, answers:Sequence[Message]) ->Tuple[Message, Union[Tile, 'Meld', None]]:
         """this is where the robot AI should go.
         Returns answer and one parameter"""
-        # pylint: disable=too-many-branches
         # disable warning about too many branches
-        answer = parameter = None
+        assert self.player.game
+        answer:Message = answers[0]
+        parameter:Any = None
         tryAnswers = (
             x for x in [Message.MahJongg, Message.OriginalCall, Message.Kong,
                         Message.Pung, Message.Chow, Message.Discard] if x in answers)
@@ -234,11 +249,10 @@ class AIDefaultAI:
         discard = self.player.game.lastDiscard
         if discard:
             for rule in self.player.game.ruleset.filterRules('claimness'):
-                claimness += rule.claimness(hand, discard)
+                claimness += rule.claimness(hand, discard)  # type:ignore[attr-defined]
                 if Debug.robotAI:
                     hand.debug(
-                        '%s: claimness in selectAnswer:%s' %
-                        (rule.name, claimness))
+                        f'{rule.name}: claimness in selectAnswer:{claimness}')
         for tryAnswer in tryAnswers:
             parameter = self.player.sayable[tryAnswer]
             if not parameter:
@@ -250,17 +264,15 @@ class AIDefaultAI:
             elif tryAnswer == Message.Pung and self.player.maybeDangerous(tryAnswer):
                 continue
             elif tryAnswer == Message.Chow:
-                parameter = self.selectChow(parameter)
+                parameter = self.selectChow(cast('MeldList', parameter))
             elif tryAnswer == Message.Kong:
-                parameter = self.selectKong(parameter)
+                parameter = self.selectKong(cast('MeldList', parameter))
             if parameter:
                 answer = tryAnswer
                 break
-        if not answer:
-            answer = answers[0]  # for now always return default answer
         return answer, parameter
 
-    def selectChow(self, chows):
+    def selectChow(self, chows:'MeldList') ->Optional['Meld']:
         """selects a chow to be completed. Add more AI here."""
         for chow in chows:
             # a robot should never play dangerous
@@ -276,14 +288,14 @@ class AIDefaultAI:
                         return chow
         return None
 
-    def selectKong(self, kongs):
+    def selectKong(self, kongs:'MeldList') ->Optional['Meld']:
         """selects a kong to be declared. Having more than one undeclared kong is quite improbable"""
         for kong in kongs:
             if not self.player.mustPlayDangerous(kong):
                 return kong
         return None
 
-    def handValue(self, hand):
+    def handValue(self, hand:'Hand') ->int:
         """compute the value of a hand.
         This is not just its current score but also
         what possibilities to evolve it has. E.g. if
@@ -322,13 +334,15 @@ class AIDefaultAI:
 #        return result
 
 
-class TileAI(StrMixin):
+class TileAI(ReprMixin):
 
     """holds a few AI related tile properties"""
     # pylint: disable=too-many-instance-attributes
     # we do want that many instance attributes
 
-    def __init__(self, candidates, tile):
+    def __init__(self, candidates:'DiscardCandidates', tile:Tile) ->None:
+        assert candidates.player
+        assert candidates.player.game
         self.tile = tile
         self.group, self.value = tile.group, tile.value
         if tile.isReal:
@@ -352,13 +366,13 @@ class TileAI(StrMixin):
         self.prev2 = None
         self.next2 = None
 
-    def __lt__(self, other):
+    def __lt__(self, other:object) ->bool:
         """for sorting"""
-        return self.tile < other.tile
+        return self.tile < cast('TileAI', other).tile
 
-    def __str__(self):
-        dang = ' dang:%d' % self.dangerous if self.dangerous else ''
-        return '%s:=%.4f%s' % (self.tile, self.keep, dang)
+    def __str__(self) ->str:
+        dang = f' dang:{int(self.dangerous)}' if self.dangerous else ''
+        return f'{self.tile}:={self.keep:.4f}{dang}'
 
 
 class DiscardCandidates(list):
@@ -366,37 +380,41 @@ class DiscardCandidates(list):
     """a list of TileAI objects. This class should only hold
     AI neutral methods"""
 
-    def __init__(self, player, hand):
+    def __init__(self, player:'PlayingPlayer', hand:'Hand') ->None:
         list.__init__(self)
         self._player = weakref.ref(player)
         self._hand = weakref.ref(hand)
         if Debug.robotAI:
-            player.game.debug('DiscardCandidates for hand %s are %s' % (
-                hand, hand.tilesInHand))
-        self.hiddenTiles = [x.exposed for x in hand.tilesInHand]
-        self.groupCounts = IntDict()
+            assert player.game
+            player.game.debug(f'DiscardCandidates for hand {hand} are {hand.tilesInHand}')
+        self.hiddenTiles:List[Tile] = [x.exposed for x in hand.tilesInHand]
+        self.groupCounts:Dict[str, int] = IntDict()
                                    # counts for tile groups (sbcdw), exposed
                                    # and concealed
         for tile in self.hiddenTiles:
             self.groupCounts[tile.group] += 1
-        self.declaredGroupCounts = IntDict()
-        for tile in sum((x for x in hand.declaredMelds), []):
+        self.declaredGroupCounts:Dict[str, int] = IntDict()
+        for tile in chain(*hand.declaredMelds):
             self.groupCounts[tile.lowerGroup] += 1
             self.declaredGroupCounts[tile.lowerGroup] += 1
         self.extend(TileAI(self, x) for x in sorted(set(self.hiddenTiles)))
         self.link()
 
     @property
-    def player(self):
+    def player(self) ->'PlayingPlayer':
         """hide weakref"""
-        return self._player() if self._player else None
+        result = self._player()
+        assert result
+        return result
 
     @property
-    def hand(self):
+    def hand(self) ->'Hand':
         """hide weakref"""
-        return self._hand() if self._hand else None
+        result = self._hand()
+        assert result
+        return result
 
-    def link(self):
+    def link(self) ->None:
         """define values for candidate.prev and candidate.next"""
         prev = prev2 = None
         for this in self:
@@ -428,7 +446,7 @@ class DiscardCandidates(list):
                 if not this.next2:
                     this.next2 = TileAI(self, this.next.tile.nextForChow)
 
-    def unlink(self):
+    def unlink(self) ->None:
         """remove links between elements. This helps garbage collection."""
         for this in self:
             this.prev = None
@@ -436,14 +454,14 @@ class DiscardCandidates(list):
             this.prev2 = None
             this.next2 = None
 
-    def best(self):
+    def best(self) ->Tile:
         """return the candidate with the lowest value"""
         lowest = min(x.keep for x in self)
         candidates = sorted(x for x in self if x.keep == lowest)
+        assert self.player.game
         result = self.player.game.randomGenerator.choice(
             candidates).tile.concealed
         if Debug.robotAI:
             self.player.game.debug(
-                '%s: discards %s out of %s' %
-                (self.player, result, ' '.join(str(x) for x in self)))
+                f"{self.player}: discards {result} out of {' '.join(str(x) for x in self)}")
         return result
